#include <u.h>
#include <libc.h>
#include <auth.h>
#include <ctype.h>
#include <bio.h>
#include <ndb.h>

#include "httpd.h"
#include "httpsrv.h"

#define setenv(var, value) if(value) putenv(var, value)
#define debug if

typedef struct Endpoints	Endpoints;

struct Endpoints
{
	char	*lsys;
	char	*lserv;
	char	*rsys;
	char	*rserv;
};

typedef struct Handle Handle;
struct Handle
{
	char *mimetype;
	char **argv;
	int	ramfs;
};

/*	our default options	*/
char *rewrite = "/sys/lib/httpd.rewrite";
char *webroot = "/usr/web";
char *HTTPLOG = "http";
static char *namespace = "/lib/namespace.httpd";
static char *address = "tcp!*!http";
static int	server = 0; // 0: runs under the control of listen. 1: runs in server mode.
static int timeout = 5; // timeout for CGI
static int mntok=0;
static char *netdir = "/net";
static char *runas="none";
static int usermode = 0; // 1: httpd service by the user who invoked httpd
static int monmode = 0; // 1: not run as daemon; monitored by `mon'
static int noweb = 0; // allows $home/web as HomePage if present

/* internal params */
char *serversoft="Pegasus/1.0 (Plan9)";
static const int contmax = 100; // max continuation count for safty

/* used internally for control */
char *myport; //  80 if you use default http port. use also in log.c
static char *myip; //  202.250.160.122 for example
static char *myname; // example: ar in ar.aichi-u.ac.jp
static char *mydom; // example: aichi-u.ac.jp
static int cpid; // pid of CGI
static int mnt; // namespace configuration is finished
static int cont; // Keep-Alive continuation count
static int notestatus; // alarm note to CGI is rceived
static char *webmnt; // to be "/usr/web/mnt"

/*	Note on cont
*	cont is used as a flag so that we can detect re-request using Keep-Alive
*	if (cont > 0)
*		re-request due to Kepp-Alive
*		so we should not reconfigure our namespace
*		cont means continued
*/

static	void		becomenone(void);
static	char*	csquery(char*, char*, char*);
static	void		dolisten(char*);
static	int		doreq(HConnect*);
static	int		send(HConnect*);
static	char*		sysdom(void);
static	Endpoints*	getendpoints(char*);
static 	void		doservice(Endpoints *end);
static 	int		execute(HConnect *c, char **argv, int ramfs);
static	int		executable(char *path);
static	int		waitfor(int pid, char *msg);
static	int		hparse2env(HConnect *c, char *head);
static	char*	trim(char *s);
static 	int		notefun(void *a, char *msg);
static 	int	execsafely(HConnect *c, char *path, char **argv, int ramfs);
static	int		merge(char *mnt, char *root);
static	int	ipcheck(HConnect *c, char *file, char *ip, char *path);
static 	char*	redirect(HConnect *c, char *file, char *path);
static 	Handle*	handler(HConnect *c, char *file, char *path);
static	void		setcgienv(HConnect *c);
static 	char *	hcat2(HConnect *c, char *a1, char *a2);
static 	char *	skip(char *s, char *p);

char *usage=
	"usage: httpd [-Mmsuw] [-t timeout] [-a srvaddress] [-d domain] [-n namespace] [-r webroot] [-R rewrite]  [[httpd] dummy ..  incalldir]\n";

void
main(int argc, char **argv)
{
	char buf[256];
	char *incalldir = nil;
	hmydomain = nil;
	fmtinstall('D', hdateconv);
	fmtinstall('H', httpconv);
	fmtinstall('U', hurlconv);
	ARGBEGIN{
	case 'n':
		namespace = ARGF();
		break;
	case 'a':
		address = ARGF(); /* tcp!*!http */
		break;
	case 'd':
		hmydomain = ARGF();
		break;
	case 't':
		timeout = atoi(ARGF());
		break;
	case 'r':
		webroot = ARGF();
		break;
	case 'R':
		rewrite = ARGF();
		break;
	case 's':
		server = 1;
		break;
	case 'm':
		mntok = 1;
		break;
	case 'u':
		usermode = 1;
		break;
	case 'w':
		noweb = 1;
		break;
	case 'M':
		monmode = 1;
		break;
	default:
		sysfatal(usage);
		break;
	}ARGEND

	if(argc > 0)
		incalldir = argv[argc-1];
	if((server && incalldir) || (!server && !incalldir))
		sysfatal("# error: inconsistent use of -s and incalldir: %s", usage);
	if(!server && monmode)
		sysfatal("# error: inconsistent use of -M: %s", usage);

	if(webroot[0] != '/')
		sysfatal("webroot must begin with '/'");
	
	snprint(buf, sizeof buf, "%s/mnt", webroot);
	webmnt = strdup(buf);

	if(namespace[0] != '/')
		sysfatal("namespace must begin with `/'");
	if(rewrite[0] != '/')
		sysfatal("rewrite must begin with `/'");
	if(hmydomain == nil)
		hmydomain = sysdom();
	/* extract mydom */
	myname = sysname();
	if(strlen(hmydomain) < strlen(myname)){
		/* something wrong */
		syslog(0, HTTPLOG, "invalid sysname and sysdom: %s %s", myname, hmydomain);
		exits("sysname > sysdom");
	}
	mydom = hmydomain + strlen(myname) + 1;

	if(!server){
		Endpoints *ends;
		if(usermode)
			sysfatal("inconsistent options");
		time(nil);
		contentinit();
		urlinit();
		statsinit();
		ends = getendpoints(incalldir);
		doservice(ends);
		exits(nil);
	}

	if(!monmode){
		switch(rfork(RFNOTEG|RFPROC|RFFDG|RFNAMEG)) {
		case -1:
			sysfatal("fork");
		case 0:
			break;
		default:
			exits(nil);
		}
	}

	/*
	 * open all files we might need before castrating namespace
	 */
	time(nil);
	syslog(0, HTTPLOG, nil);

	contentinit();
	urlinit();
	statsinit();

	runas = getuser();
	if(!usermode && (strcmp(runas, "none") != 0)){
		runas = "none";
		becomenone();
	}
	dolisten(address);
	exits(nil);
}

static void
becomenone(void)
{
	int fd;

	fd = open("#c/user", OWRITE);
	if(fd < 0 || write(fd, "none", strlen("none")) < 0)
		sysfatal("can't become none");
	close(fd);
	if(newns("none", nil) < 0)
		sysfatal("can't build normal namespace");
}

static HConnect*
mkconnect(void)
{
	HConnect *c;

	c = ezalloc(sizeof(HConnect));
	c->hpos = c->header;
	c->hstop = c->header;
	c->replog = 0;
	return c;
}

static HSPriv*
mkhspriv(void)
{
	HSPriv *p;

	p = ezalloc(sizeof(HSPriv));
	return p;
}

static void
dolisten(char *address)
{
	Endpoints *ends;
	char ndir[NETPATHLEN], dir[NETPATHLEN], *p;
	int ctl, nctl, data, spotchk;
	char		netdirb[256];

	spotchk = 0;
	syslog(0, HTTPLOG, "httpd starting");
	ctl = announce(address, dir);
	if(ctl < 0){
		syslog(0, HTTPLOG, "error: can't announce on %s: %r", address);
		return;
	}

	/* 	the logic bellow seems curious to me -Kenar- */
	strcpy(netdirb, dir);
	p = nil;
	if(netdir[0] == '/'){ // netdirb !? -Kenar-
		p = strchr(netdirb+1, '/');
		if(p != nil)
			*p = '\0';
	}
	if(p == nil)
		strcpy(netdirb, "/net");
	netdir = netdirb;

	for(;;){

		/*
		 *  wait for a call (or an error)
		 */
		nctl = listen(dir, ndir);
		if(nctl < 0){
			syslog(0, HTTPLOG, "error: can't listen on %s: %r", address);
			syslog(0, HTTPLOG, "error+ ctls = %d", ctl);
			return;
		}

		/*
		 *  start a process for the service
		 */
		switch(rfork(RFFDG|RFPROC|RFNOWAIT|RFNAMEG)){
		case -1:
			close(nctl);
			continue;
		case 0:
			/*
			 *  see if we know the service requested
			 */
			data = accept(ctl, ndir);
			if(data < 0){
				syslog(0, HTTPLOG, "error: can't open %s/data: %r", ndir);
				exits(nil);
			}
			dup(data, 0);
			dup(data, 1);
			dup(data, 2);
			close(data);
			close(ctl);
			close(nctl);

			ends = getendpoints(ndir);
			doservice(ends);

			exits(nil);
		default:
			close(nctl);
			break;
		}

		if(++spotchk > 50){
			spotchk = 0;
			contentinit();
			urlinit();
			statsinit();
		}
	}
}

static void
doservice(Endpoints *ends)
{
	HSPriv *hp;
	HConnect *c;
	Ndb *db;
	Ndbs s;
	Ndbtuple *nt;
	char buf[Ndbvlen];
	int t, ok;

	c = mkconnect();
	hp = mkhspriv();
	hp->remotesys = ends->rsys;
	hp->remoteserv = ends->rserv;
	myip = ends->lsys;
	myport=ends->lserv;
	c->private = hp;

	db = ndbopen(nil);
	if(db == 0){
		syslog(0, HTTPLOG, "error: no db files");
		exits("no db");
	}
	/*	we pic up only one	*/
	nt = ndbgetval(db, &s, "ip", ends->lsys, "sys", buf); 
	if(nt)
		ndbfree(nt);
	ndbclose(db);
	if(!nt){
		syslog(0, HTTPLOG, "error: no sys name in ndb file");
		exits("no sysname");
	}
	myname = strdup(buf);

	hinit(&c->hin, 0, Hread);
	hinit(&c->hout, 1, Hwrite);

	/*
	 * serve requests until a magic request.
	 * later requests have to come quickly.
	 * only works for http/1.1 or later.
	 */
	for(t = 15*60*1000; ; t = 15*1000){
		if(hparsereq(c, t) <= 0)
			exits(nil);
		ok = doreq(c);

		if(c->head.closeit || ok < 0)
			exits(nil);

		hreqcleanup(c);
	}
}

static int
doreq(HConnect *c)
{
	char *uri, *origuri, *p, *host;
	char virtualhost[100];
	char *urioptions;

	uri = c->req.uri;
	/*
	 * normal case is just file transfer
	 */
	if(cont > 0){
		logit(c, "continue: %d uri=%s",cont, uri);
		if(cont > contmax){ /* may be required for safty */
			logit(c, "continuation exceed max: %d uri=%s",cont, uri);
			return hfail(c, HInternal);
			exits(nil);
		}
		cont++;
	}

	if(hparseheaders(c, 15*60*1000) < 0)
		exits("failed");
	if(!http11(c) && !c->head.persist)
		c->head.closeit = 1;

	{	/*	logit		*/
		char *p1,*p2;
		int n1 = 0;
		int n2 = 0;
		int n;
		p1 = p2 = nil;
		if(c->req.params)
			p1 = c->req.params;
		if(c->req.search)
			p2 = c->req.search;
		if(p1)
			n1 = strlen(p1);
		if(p2)
			n2 = strlen(p2);
		n = n1 + n2 + 3 + UTFmax;
		urioptions = halloc(c, n);
		if(p1 && p2)
			snprint(urioptions, n, ";%s?%s", p1, p2);
		else if(p1)
			snprint(urioptions, n, ";%s", p1);
		else if(p2)
			snprint(urioptions, n, "?%s", p2);
		else
			strcpy(urioptions,"");
		logit(c, "%s %s%s", c->req.meth, uri, urioptions);
	}

	/*	Let ar be a host name and aichi-u.ac.jp is the dom
	*	then head.host is:
	*	ar
	*	ar.aichi-u.ac.jp
	*	ar.aichi-u.ac.jp:80	# some browsers send GET header with :80
	*	-Kenar-	*/
	host = hstrdup(c, c->head.host);
	/*	we can remove `:80' 	here */
	p = strchr(host,':');
	if(p)
		*p = 0;

	/*	we can remove mydom part from head.host
	*	so that we can simplify management of
	*	/sys/lib/httpd.rewrite.
	*	now we don't need full name for virtual host in this file.
	*	-Kenar-	*/
	if(strlen(host)>strlen(mydom)){
		char buf[256];
		snprint(buf, sizeof buf, ".%s", mydom);
		p = host + strlen(host) - strlen(buf);
		if(strcmp(p, buf) == 0)
			*p = 0;
	}

	snprint(virtualhost, sizeof virtualhost, "http://%s/", host);
	origuri = uri;
	uri = redirect(c, rewrite, virtualhost);
	/*
	*	we can reject some faked accesses here;
	*	if uri is nil and host is not myname, we can reject.
	*	example:
	*	ar	# sysname
	*	plan9	# another name of this host ar
	*	aichi-u.ac.jp	# mydom
	*	now ar.aichi-u.ac.jp and plan9.aichi-u.ac.jp are reduced to
	*	ar and plan9 respectively.
	*	On the other hand
	*	www.faked.com
	*	is keep untached.
	*/
	if(uri == nil){
		if(strcmp(host, myname) != 0 && strcmp(host, myip) != 0){
			logit(c, "denied: header host %s", host);
			return hfail(c, HNotFound);
		}
		uri = redirect(c, rewrite, origuri);
	}

	mnt = 0;
	if(uri && *uri == '*'){
		//	This case is '*' redirect in /sys/lib/httpd.rewrite such as:
		//	http://pegasus  */usr/someone/www
		if(!cont){
			int status;
			status = bind(uri + 1, webmnt , MREPL);
			if(status < 0){
				logit(c,"error: bind %s %s # %r", uri+1,webmnt);
				return hfail(c, HNotFound);
			}
			mnt = 1;
		}
	}
	else if(uri != nil){ /* Let the client know the new address */
		char *newuri;
		int n;
		n = strlen(uri) + strlen(urioptions) + 1 + UTFmax;
		newuri = halloc(c, n);
		snprint(newuri, n, "%s%s", uri, urioptions);
		logit(c,"redirect to:  %s", newuri);
		return hmoved(c, newuri);
	}

	return send(c);
}

/*	# return values
*	+1: OK
*	-1: NG
*	-Kenar-	*/
static int
send(HConnect *c)
{
	Dir dir;
	Handle *h;
	char *p, *uripath, *abspath;
	char *uri, *newuri;
	char *path = 0;
	char *requser = 0;
	int fd, fd1, n, ok;
	char buf[256];
	int status;
	HSPriv *hp = c->private;
	USED(requser);

	/* 	Here we support http://ar/~alice/abc
		then we bind /usr/alice/web to /usr/web/mnt
		and change request document to /abc
		This redirection is automatically done without /sys/lib/httpd.rewrite
		-Kenar-
	 */
	uri = c->req.uri;
	if(noweb == 0 && mnt == 0 && uri[0]=='/' && uri[1]=='~'){
		char *t;
		t = strchr(uri+2, '/');

		/* 	if uri is "/~alice" for example,
			let the client resend the correct uri "/~alice/"
			-Kenar-	*/
		if(!t){
			snprint(buf, sizeof buf, "%s/", uri);
			return hmoved(c, buf);
		}

		if(!cont){
			requser = ezalloc(t -uri+1); // requser is "alice"
			strncpy(requser, uri+2,  t - uri - 2);
			c->req.user = requser;
			n = strlen("/usr//web") + strlen(uri) + 1 + UTFmax;
			path = halloc(c, n);
			snprint(path, n, "/usr/%s/web", requser);
			status = bind(path, webmnt, MREPL);
			if(status == -1){
				logit(c, "error: bind %s %s # %r", path, webmnt);
				return notfound(c, uri+1);
			}
			mnt = 1;
		}
		path=t;

		/* 	then c->req.uri is "/~alice/abc", and path is "/abc" 
			for the request http://ar/~alice/abc
			We need not be wary of "/~alice/../",
			it reduces to "/" as it should be.
			-Kenar-	*/
	}
	if(!path) path = uri;
	c->req.path = path;

	/* namespace reconfiguration; hide everything out of webroot */
	if(!cont && addns(runas, namespace) < 0){
		logit(c, "error: addns %s # %r", namespace);
		sysfatal("can't build httpd namespace");
	}

	if(!cont){
		if(mnt){
			char *wdoc, *wetc, *wmdoc, *wmetc;
			/*	 the two unmounts are for the guard
			*	when next two binds fail.
			*	for example, /usr/web/mnt/etc is absent	*/
			wdoc = hcat2(c, webroot, "/doc");
			wetc = hcat2(c, webroot, "/etc");
			wmdoc = hcat2(c, webroot, "/mnt/doc");
			wmetc = hcat2(c, webroot, "/mnt/etc");
			unmount(nil, wdoc);
			unmount(nil, wetc);
			bind(wmdoc, wdoc, MREPL);
			bind(wmetc, wetc, MREPL);
			unmount(nil, webmnt);
		}
		bind(webroot,"/", MREPL);
		cont = 1;
	}
	chdir("/");

	/*	Namespace configuration is completed.
	*	Now we are in the configured namespace.	*/

	if(strcmp(c->req.meth, "GET") != 0
		&& strcmp(c->req.meth, "HEAD") != 0
		&& strcmp(c->req.meth,"POST") != 0)
		return hunallowed(c, "GET, HEAD, POST");
	/*	NOTE
	*	PUT and DELETE method are supported by
	*		Mozilla/4.5
	*	but not supported by
	*		IE/6.0 nor Mozilla/6.2
	*	so it may be worthless to support these method
	*	-Kenar-	*/

	if(c->head.expectother || c->head.expectcont)
		return hfail(c, HExpectFail);

	newuri = redirect(c, "/etc/rewrite", uri);
	if(newuri){
		logit(c, "moved to: %s", newuri);
		return hmoved(c, newuri);
	}

	/*
	 * check for directory/file mismatch with trailing /,
	 * and send any redirections.
	 */
	n = strlen("/doc") + strlen(uri) + strlen("/index.html") + 1 + UTFmax;
	/* make a room to insert index.html */


	abspath = halloc(c, n);
	strcpy(abspath, "/doc");
	strcat(abspath,path);

	if(dirstat(abspath, &dir) < 0){
		logit(c, "error: dirstat %s # %r", abspath);
		return notfound(c, uri);
	}
	
	if(dir.mode & CHDIR){
		/*	if uri does not end with '/'
			then let the client resent the uri with "/"
			-Kenar-	*/
		if(uri[strlen(uri)-1] != '/'){
			close(fd);
			snprint(buf, sizeof buf, "%s/", uri);
			return hmoved(c, buf);
		}
		
		/*	here the form of uri is "path/"
			therefore p must be > 0 , and path ends with '/'
			-Kenar-		*/
		strcat(abspath, "index.html");
	}else if((p = strchr(abspath, '\0')) > abspath && p[-1] == '/'){
		/*	This condition means:
			path is a file and ends with '/' 
			for example,
				http://ar.aichi-u.ac.jp/index.html/
			Then let the client to resend
				http://ar.aichi-u.ac.jp/index.html
			-Kenar-	*/
		*strrchr(c->req.uri, '/') = '\0';
		logit(c, "%s hmoved", uri);
		return hmoved(c, uri);
	}

	/*	Let our sendfd() know the contents type to assign full pathname
	*	to req.uri
	*	-Kenar-	*/
	
	c->req.path = abspath;
	c->req.uri = abspath;

	/*
	*	access control must be the last
	*	because it is delicate to handle last slash
	*	so full path is required to evaluate
	*	-Kenar-	*/

	if(ipcheck(c, "/etc/allow", hp->remotesys, path) < 0)
		return hfail(c, HUnauth);

	ok = authcheck(c, "/etc/passwd", path);
	if(ok <= 0)
		return ok;


	/*	Now every redirection is finished and we have ready.
	*	requested path is already converted to absolute path.
	*	values to be used are:
	*	path: path to the file to send. where path is in httpd name space
	*	fd: open file descriptor of path
	*
	* 	CGI is supported. The name must be one of the form:
	*	*.html *.cgi *.http
	*	and they must be exec permitted
	*	-Kenar-
	*/

	/* we disallow files starting with `.' */
	p = strrchr(abspath,'/');
	if(p[1] == '.'){
		close(fd);
		logit(c, "%s not found", uri);
		return notfound(c, uri);
	}

	uripath = abspath + strlen("/doc");

	/*	call exechandler	*/
	h = handler(c, "/etc/handler", uripath);

	if(h){
		char **argv;
		int n;
		int ok;
		argv = h->argv;
		ok = 0;
		for(n = 0; argv[n]; n++)
			if(strcmp(argv[n],"$target") == 0){
				argv[n] = abspath;
				ok = 1;
			}
		if(!ok){
			argv[n++] = abspath;
			argv[n] = nil;
		}

		if(*argv[0] != '/'){
			logit(c, "/etc/handler: %s must be absolute path", argv[0]);
			return hfail(c, HInternal);
		}
		if(executable(argv[0])){
			/*	Don't use access(path, AEXEC)==0 instead of executable(path)
			*	Why? If the fs is running in allow mode, all files are
			*		access(path, AEXEC)==0
			*	-Kenar-	*/
			if(strcmp(c->req.meth,"POST") == 0){
				int fd[2];
				char *hb;
				char *t;
				Hio *hin;
		
				/* ref: wiki.c */
				hb = nil;
				hin = hbodypush(&c->hin, c->head.contlen, c->head.transenc);
				if(hin != nil){
					alarm(15*60*1000); // why so much time ? -Kenar-
					hb = hreadbuf(hin, hin->pos);
					alarm(0);
				}
				if(hb == nil){
					hfail(c, HBadReq);
					exits("failed");
				}
				/* pass the hb to exec() */
				pipe(fd);
				switch(fork()){
				case -1:
					logit(c, "error: fork # %r");
					hfail(c, HInternal);
					return -1;
				case 0:
					/*	Let the child exit, because I would like to let CGI
					*	succeed to this pid
					*	-Kenar-	*/
					close(0);
					close(fd[0]);
					dup(fd[1], 1);
					close(fd[1]);
					/*	POST data have been received in hb,
					*	so we pass the data to the child process */
					write(1,hb, c->head.contlen);
					exits(nil);
				default:
					close(fd[1]);
					dup(fd[0], 0);
					close(fd[0]);
					break;
				}
			}
	
			if(strcmp(h->mimetype,"-") != 0){
				Hio *hout = &c->hout;
				hprint(hout, "%s 200 OK\r\n", hversion);
				hprint(hout, "Server: %s\r\n", serversoft);
				hprint(hout, "Date: %D\r\n", time(nil));
				hprint(hout, "Content-Type: %s\r\n", h->mimetype);
				if(c->head.closeit)
					hprint(hout, "Connection: close\r\n");
				else if(!http11(c))
					hprint(hout, "Connection: Keep-Alive\r\n");
				hprint(hout, "\r\n");
				hflush(hout);
			}
	
			setcgienv(c);

			/* chdir to the target dir, taget is abspath in httpd space*/
			p = strrchr(abspath,'/');
			*p = 0;
			chdir(abspath);
			*p = '/';
			// temporally we don't return
			//return execute(c,argv, h->ramfs);
			execute(c,argv, h->ramfs);
			exits(nil);
		}
	}
	status = sendit(c, abspath, nil, nil);
	return status; // +1: success;  -1: fail
}

static char*
sysdom(void)
{
	char *dn;

	dn = csquery("sys" , sysname(), "dom");
	if(dn == nil)
		dn = "who cares";
	return dn;
}

/*
 *  query the connection server
 */
static char*
csquery(char *attr, char *val, char *rattr)
{
	char token[64+4];
	char buf[256], *p, *sp;
	int fd, n;

	if(val == nil || val[0] == 0)
		return nil;
	snprint(buf, sizeof(buf), "%s/cs", netdir);
	fd = open(buf, ORDWR);
	if(fd < 0)
		return nil;
	fprint(fd, "!%s=%s", attr, val);
	seek(fd, 0, 0);
	snprint(token, sizeof(token), "%s=", rattr);
	for(;;){
		n = read(fd, buf, sizeof(buf)-1);
		if(n <= 0)
			break;
		buf[n] = 0;
		p = strstr(buf, token);
		if(p != nil && (p == buf || *(p-1) == 0)){
			close(fd);
			sp = strchr(p, ' ');
			if(sp)
				*sp = 0;
			p = strchr(p, '=');
			if(p == nil)
				return nil;
			return estrdup(p+1);
		}
	}
	close(fd);
	return nil;
}


/*	dir is "/net/tcp/5" for example
*	file is "remote", "local"
*	 -Kenar-	*/
static void
getendpoint(char *dir, char *file, char **sysp, char **servp)
{
	int fd, n;
	char buf[128];
	char *sys, *serv;

	sys = serv = nil;

	snprint(buf, sizeof buf, "%s/%s", dir, file); 
	fd = open(buf, OREAD);
	if(fd >= 0){
		n = read(fd, buf, sizeof(buf)-1);
		/* then we will read "202.250.160.40!80" for success */
		if(n>0){
			buf[n-1] = 0;
			serv = strchr(buf, '!');
			if(serv){
				*serv++ = 0;
				serv = estrdup(serv);  // "80" for example
			}
			sys = estrdup(buf); // "202.250.160.40" for example
		}
		close(fd);
	}
	if(serv == nil)
		serv = estrdup("unknown");
	if(sys == nil)
		sys = estrdup("unknown");
	*servp = serv;
	*sysp = sys;
}

static Endpoints*
getendpoints(char *dir)
{
	Endpoints *ep;

	ep = ezalloc(sizeof(*ep));
	getendpoint(dir, "local", &ep->lsys, &ep->lserv);
	getendpoint(dir, "remote", &ep->rsys, &ep->rserv);
	return ep;
}

/*------- Kenar's addition follows --------*/

static void
setcgienv(HConnect *c)
{
	HSPriv *hp = c->private;
	char buf[256];
	char *cputype, *name;
	int ch;

	cputype = getenv("cputype");

	rfork(RFCENVG);

	/*	Putting header info into environment
	*	must be the first. (for security)
	*	-Kenar-	*/
	ch = *(c->hstop);
	*(c->hstop) = 0;
	setenv("HTTP_HEADER", (char*)c->header);
	hparse2env(c, (char*)c->header);
	*(c->hstop) = ch;

	setenv("DOCUMENT_ROOT", "/doc");
	setenv("GATEWAY_INTERFACE", "CGI/1.1");

	if(!getenv("CONNECTION"))
		putenv("CONNECTION", "Keep-Alive");

	snprint(buf, sizeof buf, "%s!%s", hp->remotesys, hp->remoteserv);
	setenv("REMOTE_ADDR", buf);

	setenv("REQUEST_PATH", c->req.path);
	setenv("REQUEST_METHOD",c->req.meth);
	
	snprint(buf, sizeof buf, "%d.%d", c->req.vermaj,c->req.vermin);
	setenv("REQUEST_VERSION", buf);
	setenv("REQUEST_URI",c->req.path + strlen("/doc"));
	setenv("AUTH_USER", c->head.authuser);
	setenv("AUTH_PASS", c->head.authpass);
	setenv("REQUEST_USER", c->req.user);

	setenv("SERVER_DOMAIN", hmydomain);
	setenv("SERVER_NAME", sysname());
	setenv("SERVER_PORT", myport);
	setenv("SERVER_PROTOCOL",hversion);
	setenv("SERVER_SOFTWARE", serversoft);
	setenv("QUERY_STRING", c->req.search);
	setenv("target", c->req.path);
	name = strrchr(c->req.path, '/');
	setenv("name", name+1);
	setenv("query", c->req.search);
	snprint(buf, sizeof buf, "%D", time(nil));
	setenv("date", buf);
	setenv("home", "/doc");
	setenv("cputype", cputype);
	setenv("objtype", cputype);
	snprint(buf, sizeof buf, "%d", timeout);
	setenv("timeout", buf);
	setenv("namespace", namespace);
}

static int
execute(HConnect *c, char *hargv[], int ramfs)
{
	char *p, *q;
	char *params;
	int k, n, hargc;
	char **argv;
	char *path;

	params = c->req.params;
	/*	params points "alice;bob"
	*	In creating argv for exec call,
	*	we should allocate memory for argv.
	*	argv[0] must be reserved for basename of exec file
	*	argv[n-1] must be nil where n is num of elements
	*	so, I use ";alice;bob;" to apply getfield.
	*	-Kenar-		*/
	if(params == nil)
		argv = hargv;
	else{
		for(n = 0; hargv[n]; n++);
		hargc = n;
			
		n = strlen(";") + strlen(params) + 1 + UTFmax;
		q = halloc(c, n);
		snprint(q, n, "%s;", params);
		params = q;
		n = 0;
		for(p = params; *p; p++)
			if(*p == ';') n++;
		argv = halloc(c, (hargc + n+1) * sizeof(char*));
		n = getfields(params, &argv[hargc], n+1, 0, ";");
		for(k = 0; k < hargc; k++)
			argv[k] = hargv[k];
		/* don't forget to hurlunesc */
		for(k = hargc; k < hargc + n -1; k++)
			argv[k] = hurlunesc(c, argv[k]);
		/* it is not good idea to trim extra spaces */
		argv[hargc + n-1] = nil;
	}
	path = argv[0];
	p = strrchr(argv[0],'/');
	argv[0] = p + 1;

	/*	if it takes so much time to complete, kill the process
	*	-Kenar-	*/
	return execsafely(c, path, argv, ramfs);
}


/*	This code intends to get exit status from the child	*/
static int
execsafely(HConnect *c, char *path, char **argv, int ramfs)
{
	int fd, pid;
	int status;
	char msg[ERRLEN];

	if(ramfs){
		/*	ramfs service to the child	*/
		switch(pid = rfork(RFFDG|RFPROC)){ /* assign = */
		case -1:
			logit(c, "error: fork # %r");
			hfail(c, HInternal);
			return -1;
		case 0:
			close(0);
			close(1);
			close(2);// prevent "ramfs: unmount read" message
			execl("/bin/ramfs","ramfs",0);
			/* not reached */
			logit(c, "error: unable to execl /bin/ramfs %r");
			exits(nil);
		default:
			break;
		}
		
		/* ensure "/bin/ramfs" finished */
		status = waitfor(pid, msg);
		if(status < 0){
			logit(c,"error: /bin/ramfs: no children to wait %d", pid);
			return -1;
		}
	}

	/* 	wait child to finish.
	*	If it takes too much time, kill the child and get message  */
	switch(cpid = rfork(RFFDG|RFPROC|RFNOTEG)){ /* assign = */
	case -1:
		logit(c, "error: fork # %r");
		hfail(c, HInternal);
		return -1;
	case 0: // child: read stdin, and write to stdout
		if(!mntok){
			unmount(nil,"/net");
			status = rfork(RFNOMNT);
			if(status < 0){
				logit(c,"error: rfork nomnt # %r");
				return hfail(c, HInternal);
			}
		}
		addns(runas, "/etc/namespace.cgi");
		exec(path, argv);
		/* not reached */
		logit(c,"error: exec %s # %r", path);
		exits(nil);// don't return
		break;
	default:// parent: wait for child to finish.
		break;
	}
	fd = dup(0, -1);
	close(0);
	atnotify(notefun,1);
	alarm(timeout*1000);
	notestatus = 1;
	status = waitfor(cpid, msg);
	alarm(0);
	atnotify(notefun,0);
	dup(fd, 0);
	close(fd);
	unmount(nil,"/tmp");
	if(notestatus != 1){
		logit(c,"alarm done to pid=%d", cpid);
		return -1;
	}
	if(status < 0){
		logit(c,"error: no children to wait %d", cpid);
		return -1;
	}
	if(msg[0] != 0){
		logit(c,"child %d exits with msg %s", cpid, msg);
		return -1;
	}
	logit(c, "done");
	return 1;
}


/*
*	exeecutable returns 1
*	if the file is executable for `others'
*	else returns 0
*/
static int
executable(char *path)
{
	Dir dir;
	dirstat(path, &dir);
	if(dir.mode & CHEXEC)
		return 1;
	return 0;
}

static int
waitfor(int pid, char *msg)
{	int cpid;
	Waitmsg w;
	while((cpid = wait(&w)) >= 0){
		if(pid == cpid){
			strncpy(msg, w.msg, ERRLEN);
			return 0;
		}
	}
	return -1;
}


/*	mime parse is not easy.
*	and probably meaningless to put the info int env.
*	so I chose a simple way
*	I parsed only
*		key:val
*	pairs, and
*		putenv(key, val)
*	-Kenar-	*/
static int
hparse2env(HConnect *c, char *head)
{
	char *p,*s;
	char **buf;
	char *kv[2];
	char buf1[64];
	int k,n,m;

	p = hstrdup(c, head);
	/* skip a line */
	p = strchr(p, '\n');
	if(!p)
		return -1;
	p++;
	n = 0;
	for(s = p; *s; s++)
		if(*s == '\n') n++;
	buf = halloc(c, (n + 1) * sizeof(char*));
	n = getfields(p, buf, n + 1, 0, "\n");
	
	/* strip the trailing CR */
	for(k = 0; k < n; k++){
		p = strchr(buf[k], '\r');
		if(p)
			*p = 0;
	}

	/*	get
	*		key: val
	*	and trim, then putenv	*/
	for(k = 0; k < n; k++){
		m = getfields(buf[k], kv, 2, 0, ":");
		if(m != 2)
			continue;
		for(p = kv[0]; *p; p++)
			*p = toupper(*p);
		for(p = kv[0]; *p; p++)
			if(*p == '-') *p = '_';
		snprint(buf1, sizeof buf1, "HTTP_%s", trim(kv[0]));
		setenv(buf1, trim(kv[1]));
	}
	return 0;
}


static char*
trim(char *s)
{
	char *t;
	while(isspace(*s))
		s++;
	t = strchr(s, 0);
	t--;
	while(isspace(*t ))
		t--;
	t++;
	*t = 0;
	return s;
}

/*	"/proc" is required for postnote	*/

static int
notefun(void *a, char *msg)
{
	char buf[32];
	int fd;
	USED(a);
 	postnote(PNGROUP, cpid, msg);
	if(strcmp(msg, "alarm") == 0){
		notestatus = 0;
		return 1;
	}
	return 0;
}

#ifdef nodef
*	shell style path matching function
*	I modified to simplify original one
*	and made some modification of the matching rule.
*	now, '/' is a simplly regular charactor except
*	"/*/" is match to "/".
*	ref: /sys/src/cmd/rc/glob.c
*	-Kenar-
*
 * 	Does the string s match the pattern p
 * 	* matches any sequence of characters
 * 	? matches any single character
 * 	[...] matches the enclosed list of characters
 *	~ in the beginning of [...] means compliment
 *	- in [...] means range in ascii order
 *
 *	return value
 *	1: matched
 *	0: not matched
 #endif
static int
match(char *s, char *p)
{
	char *s0;
	int compl, hit, lo, hi, t, c;
	s0 = s;
	for(;*p; s++,p++){
		switch(*p){
		case '*':
			++p;
			if(*p == '/' && s > s0)
				if(match(s-1, p)) return 1;
			for(;;){
				if(match(s, p)) return 1;
				if(!*s) break;
				s++;
			}
			return 0;
		case '?':
			if(*s=='\0') return 0;
			break;
		case '[':
			if(*s=='\0') return 0;
			c=*s;
			p++;
			compl=*p=='~';
			if(compl) p++;
			hit=0;
			while(*p!=']'){
				if(*p=='\0') return 0;		/* syntax error */
				lo=*p;
				p++;
				if(*p!='-') hi=lo;
				else{
					p++;
					if(*p=='\0') return 0;	/* syntax error */
					hi=*p;
					p++;
					if(hi<lo){ t=lo; lo=hi; hi=t; }
				}
				if(lo<=c && c<=hi) hit=1;
			}
			if(compl) hit=!hit;
			if(!hit) return 0;
			break;
		default:
			if(*p != *s) return 0;
		}
	}
	return *s=='\0';
}


static char *
skip(char *s, char *p)
{
	while(*s && strchr(p, *s))
		s++;
	return s;
}

static char *
skipto(char *s, char *p)
{
	while(*s && !strchr(p, *s))
		s++;
	return s;
}

/*	getword get a word from s and returns point
*	next to search ( the point is next char of the word)
*	if there is no word, it returns nil
*	-Kenar-	*/
static char *
getword(char *s, char *w, int size)
{
	char *r;
	r = w + size;
	s = skip(s, " \t");
	if(*s == 0)
		return nil;
	while(*s && !strchr(" \t", *s)){
		if(w == r){
			w--;
			break;
		}
		*w++ = *s++;
	}
	*w = 0;
	return s;
}

/*
*	The format of "/etc/allow" is shown using example:
*	/lecture/for/our/students
*		192.168.[1-8].*
*		...
*	path
*		...
*	where ip is given by shell style pattern matching.
*
*	restrictions := restriction restriction ...
*	restriction := pathline iplines
*	pathline := path
*	iplines := ipline ipline ..
*	ipline := spaces ippattern ippattern ..
*
*	# return values
*	0: the IP is allowed
*	-1: the IP is disallowed
*/
static int
ipcheck(HConnect *c, char *file, char *ip, char *path)
{
	char *p, *line, *s;
	int ok, check, m;
	Biobuf *b;
	char word[64];

	//logit(c,"ipcheck: ip=%s", ip);

	b = Bopen(file, OREAD);
	if(b){
		ok = 1;
		check = 0;
		while((line = Brdline(b, '\n')) != nil){
			line[Blinelen(b)-1] = 0;
			s = strchr(line, '#');
			if(s != nil)
				*s = '\0'; 	/* chop comment */
			/* skip null line */
			if(*skip(line," \t") == 0)
				continue;

			if(strchr(" \t", line[0])){	// ip pattern of lpath
				if(check == 0) // skip the ip pattern
					continue;
				if(ok)
					continue;
				/* 	how can I get single pattern ?
				*	ip pattern has no space	OK */
				while(line = getword(line, word, sizeof word)){ // asign =
					//logit(c,"ipcheck: word=%s", word);
					if(word[0] == '~' && match(ip, word+1) == 0){
						ok = 1;
						break;
					}
					if(match(ip, word)){
						ok = 1;
						break;
					}
				}
				if(!ok) // examine more ip line
					continue;
			} else {	// path
				/* first we must check previous analysys */
				if(check && !ok)
					break;
				/* start new path line */
				p = strpbrk(line," \t");
				if(p)
					*p = 0;
				/* 	check some human error, if error put some
				*	guards and let that know. */
				if(line[0] != '/'){
					logit(c, "warning: path in /etc/allow must begin with '/': %s",line);
					path++;
				}
				m = strlen(line);
				if(strncmp(line, path, m) == 0
					&& (path[m] == 0 || path[m] == '/' || line[m-1] == '/')){
					/* 	then we must confirm
					*	the client IP is allowed to see our data */
					//logit(c,"matched: line=%s path=%s", line, path);
					check = 1;
					ok = 0;
				}
				else{
					check = 0;
					ok = 1;
				}
			}
		}
		Bterm(b);
		return  ok - 1;
	}
	return 0;
}

/*	path redirection for users and virtual hosts
*	this redirection is used after namespace configuration
*
*	# return values
*	nil: not redirected
*	else: new uri
*	-Kenar-	*/
static char*
redirect(HConnect *c, char *file, char *path)
{
	char *pat, *rep, *line, *s, *buf, *requser;
	Biobuf *b;
	char *field[2];
	int m, m0, bufsize;

	b = Bopen(file, OREAD);
	if(b){
		pat = nil;
		rep = nil;
		m0 = 0;
		while((line = Brdline(b, '\n')) != nil){
			line[Blinelen(b)-1] = 0;
			s = strchr(line, '#');
			if(s != nil)
				*s = '\0'; 	/* chop comment */
			if(getfields(line, field, 2, 1, " \t") != 2)
				continue;
			m = strlen(field[0]);
			if(m>=m0
				&& strncmp(field[0], path, m) == 0
				&& (path[m] == 0 || path[m] == '/' || field[0][m-1] == '/')){
				/*	what is this condition ? consider the case:
				*	field[0] == "/somepath" and path == "/somepathogen"
				*	we must exclude this case as matched.
				*	on the other hands what should we do on next case ?
				*	field[0] == "/somepath/" and path == "/somepath/file"
				*	this should be OK
				*	-Kenar-	*/
				if(m == m0)
					logit(c, "warning: file %s: same pattern %s, skipped", file, field[0]);
				else{
					m0 = m;
					pat = field[0];
					rep = field[1];
				}
			}
		}
		if(m0 > 0){
			if(rep[0] == '/' ) {
				/* must be a path name of this httpd name space */
				requser = c->req.user;
				if(requser){
					bufsize = strlen(requser) + strlen(rep)
						+ strlen(path) - strlen(pat) + strlen("/~") + 1 + UTFmax;
					buf = halloc(c, bufsize);
					snprint(buf, bufsize, "/~%s%s%s", requser, rep, path + strlen(pat));
				} else {
					bufsize = strlen(rep)
						+ strlen(path) - strlen(pat) + 1 + UTFmax;
					buf = halloc(c, bufsize);
					snprint(buf, bufsize, "%s%s", rep, path + strlen(pat));
				}
			}
			else{ // assumed "http://...."
				bufsize = strlen(rep) + strlen(path) - strlen(pat) + 1 + UTFmax;
				buf = halloc(c, bufsize);
				snprint(buf, bufsize, "%s%s", rep, path + strlen(pat));
			}
			rep = buf;
		}
		Bterm(b);
		return rep;
	}
	return nil;
}

/*	exec handler
*
*	# data format:
*	path	mimetype	ramfs	path	args ...
*
*	# example:
*	.dx_html	text/html	0	/bin/dx
*
*	# return values
*	nil: need not to handle
*	else: handling command line
*	-Kenar-	*/
static Handle*
handler(HConnect *c, char *file, char *path)
{
	char *line, *s, *p;
	Biobuf *b;
	char *field[16+1], **v;
	Handle *h;
	int n, ok;

	if(!path)
		return nil;
	b = Bopen(file, OREAD);
	if(b){
		ok = 0;
		p = halloc(c, 256);
		while((line = Brdline(b, '\n')) != nil){
			line[Blinelen(b)-1] = 0;
			s = strchr(line, '#');
			if(s != nil)
				*s = '\0'; 	/* chop comment */
			strncpy(p, line, 256);
			n = tokenize(p, field, 16);
			if(n < 4)
				continue;
			if(match(path, field[0])){
				ok = 1;
				field[n] = nil;
				break;
			}
		}
		Bterm(b);
		if(!ok)
			return nil;
		h = halloc(c, sizeof(Handle));
		h->mimetype = field[1];
		h->ramfs = atoi(field[2]);
		/* we must have a room for additional $target at the end */
		v = halloc(c, n * sizeof(char*));
		for(n = 3; field[n]; n++)
			v[n - 3] = field[n];
		v[n -3] = nil;
		h->argv = v;
		return h;
	}
	return nil;
}

/*	I hate interface like:
*	hcat(c, arg1, arg2, nil)
*	because I forget last nil	*/
static char *
hcat2(HConnect *c, char *a1, char *a2)
{
	int m;
	char *b;
	m = strlen(a1) + strlen(a2) + 1 + UTFmax;
	b = halloc(c, m);
	strcpy(b, a1);
	strcat(b, a2);
	return b;
}
