#include <u.h>
#include <libc.h>
#include <ctype.h>

//Maxenvsize = 16300;// see /sys/src/9/port/devenv.c

static char *reade(int fd, char *bp, char *ebp);
static char *readeln(int fd, char *bp, char *ebp);
static int	addenv(char *name, char *value);
static char* 	plus2space(char *s);
static int 		qsdec(char *qs);
static char *	urlunesc(char *s);
static char*	shapeup(char *s);
static int rflag = 0;

void main(int argc, char *argv[])
{
	char buf[1024*16], *bp, *s, *ebp;
	int n;
	int size = 0;
	ARGBEGIN{
	case 'r':
		rflag = 1;
		break;
	case 'b':
		size = atoi(ARGF());
		break;
	default: 
		sysfatal("usage: qsparse [-r] [-b bufsize] [query]");
	}ARGEND

	if(argv[0] == nil){
		bp = buf;
		ebp = buf + sizeof buf;
		if(size){
			bp = malloc(size*1024);
			if(bp == nil)
				sysfatal("no memory: %r");
			ebp = bp + size;
		}
		s = readeln(0, bp, ebp);
		if(s == nil)
			sysfatal("read error");
		if(s == ebp)
			sysfatal("buffer exhausted");
		*s = 0;
		argv[0] = bp;
	}
	s = strchr(argv[0],'\n');
	if(s)
		*s = 0;
	n = qsdec(argv[0]);
	if(n > 0)
		sysfatal("trancated");
	exits(nil);
}

static char *
reade(int fd, char *bp, char *ebp)
{
	int n;
	n = read(fd, bp, ebp - bp);
	if(n < 0)
		return nil;
	return bp + n;
}

/*
*	readeln: read chars until '\n'
*
*	return values:
*	case	nil: read error
*	case	bp == ebp: buffer overflow
*	else: then *bp == '\n'
*/
static char *
readeln(int fd, char *bp, char *ebp)
{
	char *s, *t;
	s = bp;
	while(s < ebp){
		t = reade(fd, s, ebp);
		if(t == nil)
			return nil;
		if(t == s)
			break;
		while(s < t && *s != '\n')
			s++;
		if(*s == '\n')
			break;
		s = t;
	}
	return s;
}

/*
*	the step is:
*	1.	&
*	2.	=
*	3.	+
*	4.	%XX
*
*	return values:
*		0: normal end
*		-1: no =
*		> 0 : trancated
*	-Kenar-
*/
static int
qsdec(char *qs)
{
	char **q;
	char *p,*s, *t;
	char buf[NAMELEN];
	int n, i;
	int tranc = 0; // trancation count
	/*	count '&' in the string	*/

	n = 0;
	for(s = qs; *s; s++)
		if(*s == '&')
			n++;
	q = calloc(n + 1,sizeof(char*));
	/*	getfields overwrite the string, so I must ...	*/
	s = strdup(qs);
	n = getfields(s, q, n+1, 0, "&");
	for(i = 0; i < n; i++){
		p = strchr(q[i], '=');
		if(p){
			*p++ = 0;
			s = q[i];
			/* translate and check the the value before put env */
			s = plus2space(s);
			p = plus2space(p);
			s = urlunesc(s);
			p = urlunesc(p);
			for(t= s; *t; t++)
				if(!isalnum(*t) && !strchr("_", *t))
					return -1;
			if(rflag)
				p = shapeup(p);
			snprint(buf, sizeof buf, "QS_%s", s);
			if(addenv(buf, p) < 0)
				tranc++;
		}
		else
			return -1;
	}
	return tranc;
}

/*	memo
term% a=(alice bob)
term% xd -c /env/a
0000000   a  l  i  c  e 00  b  o  b 00
000000a 
*/
static int
addenv(char *name, char *value)
{
	int fd, n, m;
	char buf[64];
	snprint(buf, sizeof buf, "/env/%s", name);
	fd = open(buf, OWRITE);
	if(fd < 0)
		fd = create(buf, OWRITE, 0666);
	if(fd < 0)
		return -1;
	seek(fd, 0, 2);
	m = strlen(value) + 1;
	n = write(fd, value, m);
	close(fd);
	if(n != m)
		return -1;
	return 0;
}


char *
plus2space(char *s)
{
	char *p;
	for(p = s; *p; p++)
		if(*p == '+')
			*p = ' ';
	return s;
}


static char *
urlunesc(char *s)
{
	char *t, *v, *u;
	Rune r;
	int c, n;

	/* unescape */
	u = malloc(strlen(s)+1);
	for(t = u; c = *s; s++){
		if(c == '%'){
			n = s[1];
			if(n >= '0' && n <= '9')
				n = n - '0';
			else if(n >= 'A' && n <= 'F')
				n = n - 'A' + 10;
			else if(n >= 'a' && n <= 'f')
				n = n - 'a' + 10;
			else
				break;
			r = n;
			n = s[2];
			if(n >= '0' && n <= '9')
				n = n - '0';
			else if(n >= 'A' && n <= 'F')
				n = n - 'A' + 10;
			else if(n >= 'a' && n <= 'f')
				n = n - 'a' + 10;
			else
				break;
			s += 2;
			c = r*16+n;
		}
		*t++ = c;
	}
	*t = 0;

	return u;
}

static char*
shapeup(char *s)
{
	char *s0,*t;
	s0 = t = s;
	while(*s){
		if(*s == '\r')
			s++;
		*t++ = *s++;
	}
	*t = 0;
	return s0;
}
