#include <u.h>
#include <libc.h>
#include <auth.h>

static char *LOG = "mon";
static char *usage="usage: mon [-d] [-u user] [-p passwdfile] cmdpath argument ...";

static char *user=nil;
static char *password=nil;
static char *passwdfile=nil;
static int daemon = 0;

static void	execv0(char *argv[]);
static int	waitfor(int pid, char *msg);
static void	becomenone(void);

void
main(int argc, char **argv)
{
	int n, t0, t, pid;
	char buf[256], msg[ERRLEN], *p;
	char *user0; // who invoked

	ARGBEGIN{
	case 'u':
		user = ARGF();
		break;
	case 'p':
		passwdfile = ARGF();
		break;
	case 'd':
		daemon = 1;
		break;
	default:
		sysfatal("%s", usage);
	}ARGEND

	if(*argv == nil)
		sysfatal("%s", usage);
	user0 = getuser();
	if(user && strcmp(user,user0) != 0){
		int fd, n;
		char *p;
		if(!passwdfile)
			sysfatal("no passwd file given");
		fd = open(passwdfile, OREAD);
		if(fd < 0)
			sysfatal("no %s: %r", buf);
		n = read(fd, buf, sizeof buf);
		buf[n] = 0;
		p = strchr(buf, '\n');
		if(p)
			*p = 0;
		close(fd);
		password = buf;
	}

	if(daemon){
		switch(rfork(RFNOTEG|RFPROC|RFFDG|RFNAMEG)) {
		case -1:
			sysfatal("fork");
		case 0:
			break;
		default:
			exits(nil);
		}
	}
	/*
	 * open all files we might need before castrating namespace
	 */
	time(nil);
	syslog(0, LOG, nil);
	syslog(0, LOG, "mon start %s by %s run as %s", *argv, user0, user);

	n = 0;
	t0 = time(nil);
	while(n < 3){
		switch(pid = fork()){// assign =
		case -1:
			sysfatal("fork: %r");
		case 0:
			if(user){
				if(strcmp(user, user0) != 0)
					login(user, password, nil);
				}
			else
				becomenone();
			execv0(argv);
			exits(nil);
		default:
			syslog(0, LOG, "waiting %d",pid);
			waitfor(pid, msg);
			if(*msg){
				p = strchr(msg,':'); p = strchr(p, ':'); p++;
				syslog(0,LOG,"%d:%s", pid, p);
			}
			else
				syslog(0,LOG,"terminated");
		}
		t = time(nil);
		if(t -t0 < 5) // something wrong
			n++;
		else
			n = 0;
		t0 = t;
	}
}

static void
becomenone(void)
{
	int fd;

	fd = open("#c/user", OWRITE);
	if(fd < 0 || write(fd, "none", strlen("none")) < 0)
		sysfatal("can't become none");
	close(fd);
	if(newns("none", nil) < 0)
		sysfatal("can't build normal namespace");
}


static void
execv0(char *argv[])
{
	char path[256], wbuf[256],*p;
	if(*argv[0] == '/')
		strcpy(path, argv[0]);
	else if(strncmp(argv[0], "./",2) == 0){
		argv[0]++;
		snprint(path, sizeof path, "%s%s", getwd(wbuf, sizeof wbuf),argv[0]);
	}
	else
		snprint(path, sizeof path, "/bin/%s", argv[0]);
	p = strrchr(path, '/');
	if(p){
		p++;
		argv[0] = p;
	}
	exec(path, argv);
	sysfatal("unable to exec: %s: %r", path);
}

static int
waitfor(int pid, char *msg)
{	int cpid;
	Waitmsg w;
	while((cpid = wait(&w)) != -1){
		if(pid == cpid){
			strncpy(msg, w.msg, ERRLEN);
			return 0;
		}
	}
	return -1;
}
